package ch.unige.solidify.controller;

import java.util.NoSuchElementException;

import javax.servlet.http.HttpServletRequest;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.web.servlet.error.ErrorAttributes;
import org.springframework.dao.CannotAcquireLockException;
import org.springframework.dao.ConcurrencyFailureException;
import org.springframework.dao.DataIntegrityViolationException;
import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.orm.ObjectOptimisticLockingFailureException;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.validation.BindException;
import org.springframework.validation.BindingResult;
import org.springframework.web.HttpRequestMethodNotSupportedException;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.annotation.ControllerAdvice;
import org.springframework.web.bind.annotation.ExceptionHandler;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.unige.solidify.exception.SolidifyHttpErrorException;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRestException;
import ch.unige.solidify.exception.SolidifyUndeletableException;
import ch.unige.solidify.exception.SolidifyUnmodifiableException;
import ch.unige.solidify.exception.SolidifyValidationException;
import ch.unige.solidify.repository.SolidifyDbLockError;
import ch.unige.solidify.repository.SolidifyError;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.validation.ValidationError;

@ControllerAdvice
public class SolidifyControllerAdvice {

  private static final String ERROR_CONTENT_TYPE = "application/hal+json";
  private static final Logger log = LoggerFactory.getLogger(SolidifyControllerAdvice.class);

  private final ErrorAttributes errorAttributes;
  private final HttpHeaders headers = new HttpHeaders();
  private final MessageService messageService;

  public SolidifyControllerAdvice(ErrorAttributes errorAttributes, MessageService messageService) {
    this.headers.add(HttpHeaders.CONTENT_TYPE, ERROR_CONTENT_TYPE);
    this.errorAttributes = errorAttributes;
    this.messageService = messageService;
  }

  /******************************/

  @ExceptionHandler(SolidifyHttpErrorException.class)
  public ResponseEntity<SolidifyError> handleException(SolidifyHttpErrorException exception, HttpServletRequest request) {
    SolidifyError solidifyError = exception.getSolidifyError();
    solidifyError.setPath(request.getRequestURI());
    return ResponseEntity.status(solidifyError.getStatus()).body(solidifyError);
  }

  @ExceptionHandler(BindException.class)
  public ResponseEntity<SolidifyError> handleException(BindException exception, HttpServletRequest request) {
    return this.buildResponse(exception, request, this.messageService.get("error.illegal.parameter"), HttpStatus.BAD_REQUEST);
  }

  /**
   * Used to catch DataIntegrityViolationException
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler(DataIntegrityViolationException.class)
  public ResponseEntity<SolidifyError> handleException(DataIntegrityViolationException exception, HttpServletRequest request) {
    final SolidifyError solidifyError = new SolidifyError(this.errorAttributes, request);
    solidifyError.setStatus(HttpStatus.BAD_REQUEST);
    solidifyError.setError("Data integrity violation exception");
    String exceptionMessage = exception.getMostSpecificCause().getMessage();
    if (exceptionMessage.contains("\n")) {
      exceptionMessage = exceptionMessage.substring(0, exceptionMessage.indexOf('\n'));
    }
    solidifyError.setMessage(exceptionMessage);
    return this.buildBadRequestResponse(exception, request, solidifyError);
  }

  /**
   * Used to catch all exceptions not specifically catched by another ExceptionHandler
   *
   * @param exception
   * @param request @returnlogException(exception, request);
   */
  @ExceptionHandler(Exception.class)
  public ResponseEntity<SolidifyError> handleException(Exception exception, HttpServletRequest request) {

    this.logException(exception, request);
    final SolidifyError error = new SolidifyError(this.errorAttributes, request);
    return ResponseEntity.status(error.getStatus()).headers(this.headers).body(error);
  }

  /**
   * Used to catch requests errors (such as invalid JSON received)
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler(HttpMessageNotReadableException.class)
  public ResponseEntity<SolidifyError> handleException(HttpMessageNotReadableException exception, HttpServletRequest request) {

    this.logException(exception, request);

    String message = this.messageService.get("error.request.notreadable");

    final Throwable nestedException = exception.getCause();
    if (nestedException instanceof JsonParseException) {
      message = this.messageService.get("error.request.json.malformed");
    }

    final SolidifyError error = new SolidifyError(this.errorAttributes, request);
    error.setMessage(message);
    error.setError(exception.getMessage());
    error.setStatus(HttpStatus.BAD_REQUEST);

    return ResponseEntity.badRequest().headers(this.headers).body(error);
  }

  @ExceptionHandler(HttpRequestMethodNotSupportedException.class)
  public ResponseEntity<SolidifyError> handleException(HttpRequestMethodNotSupportedException exception, HttpServletRequest request) {
    return this.buildResponse(exception, request, this.messageService.get("error.illegal.http.method"), HttpStatus.METHOD_NOT_ALLOWED);
  }

  @ExceptionHandler(IllegalStateException.class)
  public ResponseEntity<SolidifyError> handleException(IllegalStateException exception, HttpServletRequest request) {
    return this.buildResponse(exception, request, this.messageService.get("error.illegal.state"), HttpStatus.BAD_REQUEST);
  }

  /**
   * Used to catch Spring constraints validation errors (rules defined as annotations on entities)
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler(MethodArgumentNotValidException.class)
  public ResponseEntity<SolidifyError> handleException(MethodArgumentNotValidException exception, HttpServletRequest request) {
    final BindingResult errors = exception.getBindingResult();
    final ValidationError validationError = new ValidationError(errors);
    return this.buildBadRequestResponse(exception, request, validationError);
  }

  /**
   * Used to catch ValidationError set at the service level (such as linked resources not found)
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler(SolidifyValidationException.class)
  public ResponseEntity<SolidifyError> handleException(SolidifyValidationException exception, HttpServletRequest request) {
    return this.buildBadRequestResponse(exception, request, exception.getValidationError());
  }

  /**
   * Used to catch: AccessDeniedException.class (not enough permissions),
   * SolidifyUnmodifiableException (resource unmodifiable), SolidifyUndeletableException (resource
   * undeletable)
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler({ AccessDeniedException.class, SolidifyUnmodifiableException.class, SolidifyUndeletableException.class })
  public ResponseEntity<SolidifyError> handleForbiddenException(Exception exception, HttpServletRequest request) {
    return this.buildResponse(exception, request, this.messageService.get("error.request.forbidden"), HttpStatus.FORBIDDEN);
  }

  // Exception handler for Solidify Rest client HttpClientErrorException
  @ExceptionHandler(SolidifyRestException.class)
  public ResponseEntity<SolidifyError> handleHttpRestClientException(SolidifyRestException exception, HttpServletRequest request) {
    this.logException(exception, request);
    SolidifyError error = exception.getSourceError();
    if (error != null) {
      return ResponseEntity.status(error.getStatus()).headers(this.headers).body(error);
    } else {
      error = new SolidifyError(this.errorAttributes, request);
      return ResponseEntity.status(error.getStatus()).headers(this.headers).body(error);
    }
  }

  // Exception handler for Optimistic lock or pessimistic lock
  @ExceptionHandler({ CannotAcquireLockException.class, ObjectOptimisticLockingFailureException.class })
  public ResponseEntity<SolidifyError> handleLockException(ConcurrencyFailureException exception, HttpServletRequest request) {
    this.logException(exception, request);
    final SolidifyError error = new SolidifyDbLockError(this.errorAttributes, request, exception);
    return ResponseEntity.status(error.getStatus()).headers(this.headers).body(error);
  }

  /**
   * Used to catch NoSuchElementException/EmptyResultDataAccessException
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler({ NoSuchElementException.class, EmptyResultDataAccessException.class, SolidifyResourceNotFoundException.class })
  public ResponseEntity<SolidifyError> handleNotFoundException(Exception exception, HttpServletRequest request) {
    return this.buildResponse(exception, request, this.messageService.get("error.request.notfound"), HttpStatus.NOT_FOUND);
  }

  protected ResponseEntity<SolidifyError> buildBadRequestResponse(Exception exception, HttpServletRequest request, SolidifyError solidifyError) {
    this.logException(exception, request);
    this.logErrorAsJson(solidifyError);
    return ResponseEntity.badRequest().headers(this.headers).body(solidifyError);
  }

  protected ResponseEntity<SolidifyError> buildBadRequestResponse(Exception exception, HttpServletRequest request,
          ValidationError validationError) {

    final SolidifyError solidifyError = new SolidifyError(this.errorAttributes, request);
    solidifyError.setValidationError(validationError);
    solidifyError.setMessage(exception.getMessage());
    solidifyError.setStatus(HttpStatus.BAD_REQUEST);

    return this.buildBadRequestResponse(exception, request, solidifyError);
  }

  protected ResponseEntity<SolidifyError> buildResponse(Exception exception,
          HttpServletRequest request, String message, HttpStatus httpStatus) {
    this.logException(exception, request);

    final SolidifyError error = new SolidifyError(this.errorAttributes, request);
    error.setMessage(message);
    error.setError(exception.getMessage());
    error.setStatus(httpStatus);

    return ResponseEntity.status(httpStatus).headers(this.headers).body(error);
  }

  protected void logErrorAsJson(Object errorObject) {
    try {
      final String errorMessage = new ObjectMapper().writeValueAsString(errorObject);
      log.error(errorMessage);
    } catch (final JsonProcessingException e) {
      log.error("unable to log validation error: {}", e);
    }
  }

  protected void logException(Exception exception, HttpServletRequest request) {
    String headerXFF = request.getHeader("X-FORWARDED-FOR");
    log.error("Request url {} {} from remote address {} with X-FORWARDED-FOR {} has generated an exception caught by controller advice: {}",
            request.getMethod(), request.getRequestURL(), request.getRemoteAddr(), headerXFF, exception);
  }

}
