package ch.unige.solidify.util;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;

import javax.xml.bind.DatatypeConverter;

import org.springframework.security.crypto.encrypt.BouncyCastleAesCbcBytesEncryptor;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;

/**
 * This class provides a way to encrypt or decrypt a string. The encrypted text in Hexdecimal
 * format. Encryption and decryption use BouncyCastleAesCbcBytesEncryptor
 **/
public class StringCipherTools {

  // According to RFC 8018 salt should be at least 64 bits
  private static final int SALT_LENGTH = 8;

  /**
   * Method to decrypt text
   *
   * @param cipherText
   *          encrypted text
   * @param password
   *          password
   * @return decrypted clear text
   * @throws IOException
   */
  public static String decrypt(String cipherText, String password) {
    return decrypt(cipherText, password, SALT_LENGTH);
  }

  /**
   * Method to decrypt text
   *
   * @param cipherText
   *          encrypted text
   * @param password
   *          password
   * @param saltlength
   *          length of salt
   * @return decrypted clear text
   * @throws IOException
   */
  public static String decrypt(String cipherText, String password, int saltlength) {
    try (ByteArrayInputStream is = new ByteArrayInputStream(
            DatatypeConverter.parseHexBinary(cipherText));) {
      final byte[] salt = new byte[saltlength];
      final int cipherTextRealSize = is.available() - saltlength;
      final byte[] cipherTextPart = new byte[cipherTextRealSize];
      is.read(salt, 0, saltlength);
      is.read(cipherTextPart, 0, cipherTextRealSize);
      final BouncyCastleAesCbcBytesEncryptor encryptor = new BouncyCastleAesCbcBytesEncryptor(password,
              DatatypeConverter.printHexBinary(salt));
      return new String(encryptor.decrypt(cipherTextPart));
    } catch (final IOException e) {
      throw new SolidifyRuntimeException("Could not read buffer during decryption", e);
    }
  }

  /**
   * Method to encrypt plain text String
   *
   * @param plainText
   *          clear text to be encrypted
   * @param password
   *          password
   * @return encrypted text in Hexdecimal format
   * @throws NoSuchAlgorithmException
   */
  public static String encrypt(String plainText, String password) {
    return encrypt(plainText, password, SALT_LENGTH);
  }

  /**
   * Method to encrypt plain text String
   *
   * @param plainText
   *          clear text to be encrypted
   * @param password
   *          password
   * @param salt
   *          salt
   * @return encrypted text in Hexdecimal format
   * @throws NoSuchAlgorithmException
   */
  public static String encrypt(String plainText, String password, byte[] salt) {

    if (salt == null || salt.length == 0) {
      throw new IllegalArgumentException("Salt must not be empty");
    }

    final int saltLength = salt.length;
    final BouncyCastleAesCbcBytesEncryptor encryptor = new BouncyCastleAesCbcBytesEncryptor(password,
            DatatypeConverter.printHexBinary(salt));
    final byte[] result = encryptor.encrypt(plainText.getBytes());
    final byte[] ciphertext = new byte[saltLength + result.length];
    System.arraycopy(salt, 0, ciphertext, 0, saltLength);
    System.arraycopy(result, 0, ciphertext, saltLength, result.length);
    return DatatypeConverter.printHexBinary(ciphertext);

  }

  /**
   * Method to encrypt plain text String
   *
   * @param plainText
   *          clear text to be encrypted
   * @param password
   *          password
   * @param saltlength
   *          length of salt
   * @return encrypted text in Hexdecimal format
   * @throws NoSuchAlgorithmException
   */
  public static String encrypt(String plainText, String password, int saltlength) {

    final byte[] salt = new byte[saltlength];
    // Generate the random salt
    SecureRandom secureRandom;
    try {
      secureRandom = SecureRandom.getInstance("SHA1PRNG");
    } catch (final NoSuchAlgorithmException e) {
      throw new SolidifyRuntimeException("Algorithm SHA1PRNG not available", e);
    }
    secureRandom.nextBytes(salt);
    return encrypt(plainText, password, salt);
  }

  /**
   * This constructor should not be called
   */
  private StringCipherTools() {
    throw new IllegalStateException(SolidifyConstants.TOOL_CLASS);
  }

}
